#!/usr/bin/perl
#
# Utility to create an MP3 list file (m3u) in the track order as defined
# in the CDDB database.
# Copyright (C) 1999-2000 Louise Youdale
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# requires local cddb database, and mp3 files to only contain track name.
#
# 20 May, 2k louise@imbrugliamail.com - started

$LOCAL_CDDB = '/var/lib/cddbp/cddb';

# eoc

$| = 1;

# build list of current directory
print "Current directory contains - ";

opendir DIR, "." || die "opendir: $!";
@this_dir = grep /^.+\.mp3$/i, grep -f, readdir DIR;
closedir DIR;
$file_count = @this_dir;

# display contents
print "$file_count mp3 files\n";

exit unless ($file_count > 0);

$i = 1;
foreach $key (@this_dir) {
	print $i++ . ": $key\n";
}

# now hunt thru local cddb database for matching titles

chdir $LOCAL_CDDB || die "chdir: $!";
opendir DIR, "." || die "opendir: $!";
@all_cats = grep !/^\.\.?$/, grep -d, readdir DIR;
closedir DIR;

foreach $cat (@all_cats) {
#	print "checking category $cat\n";

	chdir $cat || die "chdir: $!";
	opendir DIR, "." || die "opendir: $!";
	@files = grep -f, readdir DIR;
	closedir DIR;

	foreach $key (@files) {
#		print "checking entry $key\n";
		parse_mp3 ($key);
	}
	chdir "..";
}

# display results

print "\n%   file name\n-------------\n";
foreach $result (reverse sort keys %result_file) {
	print int($result) . ": $result_file{$result} - $result_album{$result}\n";
}

sub parse_mp3 {
	my($file) = @_;
	
	$track_matches = 0;
	$number_tracks = 0;

	open FILE, $file || die "open: $!";
	while (<FILE>) {
		s/[\r\n]//g;
		if (/^DTITLE=(.+)/) {
			$title = $1;
		} elsif (/^TTITLE(\d+)=(.+)/) {
			$track = $2;
			$tracks{$1} = $2;
			$number_tracks++;
		}
	}
	close FILE;

# same number of tracks
	if ($number_tracks => $file_count) {
		foreach $key (@this_dir) {
			$len = length $key;

			foreach $track (keys %tracks) {
($t_key = $key) =~ s/\W+//g;
($t_track = $tracks{$track}) =~ s/\W+//g;
				if ($t_key =~ /\Q$t_track\Emp3/i) {
					$track_matches++;
					delete $lengths{$track};
				}
			}
		}

		if ($file_count eq $track_matches) {
#			print "$file: full match on \'$title\'\n";

$result = 100;
while ($result_file {$result}) {
	$result = sprintf "%3.3i", --$result;
}
$result_file {$result} = $file;
$result_album {$result} = $title;
		} elsif ($track_matches > 0) {
			$percent = int (($track_matches * 100) / $file_count);
#			print "$file: matches = $track_matches : $percent %\n";

$result = sprintf "%3.3i", $percent;
while ($result_file {$result}) {
	$result = sprintf "%3.3i", --$result;
}
$result_file {$result} = $file;
$result_album {$result} = $title;
		}
	}

# cleanup
	undef %tracks;
}

# eof
