#!/usr/bin/perl
#
# CDDBP Proxy, converts cddbp to http.
# Copyright (C) 1999-2000 Louise Youdale
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# 2 June 1999 louise@imbrugliamail.com	- released 1.0
# 5 June 1999 P.Metternich@gmx.de		- fixed category dir creation
# 11 June 1999 louise@imbrugliamail.com	- added cddb queue, changed default path
# 12 June 1999 renehl@post1.tele.dk 	- fixed cache bits
# 14 June 1999 louise@imbrugliamail.com	- remove carriage return from local files
# 4 July 1999 louise@imbrugliamail.com	- cleaner options and logging, added traces
#						  	  queue no longer quits, fixed inexact lists
# 8 October 1999 louise@imbrugliamail.com	- added max protocol level config, altered dot
#						  	  carriage return for AudioCatalyst
# 9 October 1999 louise@imbrugliamail.com	- added lscat command, and fuzzy local query
# 16 June 2k diarmuid.collins@s3group.com - altered mode of mkdir directory

# write pidfile
$PIDFILE=shift;
open( FPIDFILE, "> $PIDFILE");
print FPIDFILE $$;
close(FPIDFILE);

$ENABLE_LOG   = true;	# logging to file

$ENABLE_CACHE = true;	# on disc cddb cache
$ENABLE_FUZZY = true;	# fuzzy lookup in local cache
$ENABLE_SPOOL = true;	# store queries while offline - requires 'offline' cmd line

$LOCAL_CDDB = '/var/lib/cddb';
$CDDB_SPOOL = '/var/spool/cddbp';	# nb: spool cannot be subdir of cddb

# CLIENT = tcp to client application
# PROXY  = tcp from client application
# HTTP   = tcp from http cddb server
# CACHE  = i/o from disc cache
$TRACES = "CLIENT|PROXY|HTTP|CACHE";

# TODO: tk config

if (1) { # direct 
	$HTTP_HOST = "freedb.freedb.org";
	$HTTP_PATH = "/~cddb/cddb.cgi";
	$HTTP_PORT = "http"; # "http" or number, e.g. 80
} else { # proxy
	$HTTP_HOST = "www-cache";
	$HTTP_PATH = "http://freedb.freedb.org/~cddb/cddb.cgi";
	$HTTP_PORT = 8080;
	$FUDGE = true;
}
$CDDBP_PORT = 888; # "cddb" or number, e.g. 888

$LOWER_CASE_CDDB = true;

chomp ($hostname = `hostname`);

# eoc

$version = "v1.7";

$supported_cddb = 4;
$tolerance = 1000;

use Socket;
use Carp;

$|= 1;

print "CDDB Proxy $version (louise\@imbrugliamail.com)\n\n";

# check command line arguments

if ($ARGV[0] =~ /offline/i) {
	$SPOOL = true;
} else {
	$SPOOL = false;
}

if ($ARGV[0] =~ /h/i) {
	print <<EOD;
usage: $0 [offline]

  offline - enables write to disc query cache

configuration:

  listening on:   $hostname:$CDDBP_PORT
  remote host:    $HTTP_HOST:$HTTP_PORT
  remote path:    $HTTP_PATH
EOD

if ($ENABLE_CACHE eq true) {
	print "\n  local database: $LOCAL_CDDB";
	print " (fuzzy tolerance $tolerance)" if ($ENABLE_FUZZY eq true);
	print "\n";
}

	print "\n  spooling:       ";
	if ($ENABLE_SPOOL eq true) { print "$CDDB_SPOOL";
	} else { print "disabled"; }

	print "\n\n  logging:        to ";
	if ($ENABLE_LOG eq true) { print "file \"log\"";
	} else { print "screen"; }
	print " (trace roles $TRACES)\n";

	exit;
}

# write to disc query cache only enabled when 'offline' parameter and enable_spool are set

undef $SPOOL if ($ENABLE_SPOOL eq false);

# TODO: variable log levels

if ($ENABLE_LOG eq true) {
	open LOG, ">log" || die "log";
	$old = select LOG;
	$|= 1;
	select $old;
}

sub logmsg { 
	print     "@_\t# ", scalar localtime, " #\n";
	print LOG "@_\t# ", scalar localtime, " #\n" if ($ENABLE_LOG eq true);
}

clean_spool () if ($SPOOL eq false);
logmsg "queries will be spooled to $CDDB_SPOOL" if ($SPOOL eq true);

if ($CDDBP_PORT =~ /(\d+)/) {
	$port = $1;
} else {
	$port = getservbyname ($CDDBP_PORT, 'tcp');
}
$proto = getprotobyname ('tcp');

socket (Server, PF_INET, SOCK_STREAM, $proto) or die "socket: $!";
setsockopt (Server, SOL_SOCKET, SO_REUSEADDR, pack ("l", 1)) or die "setsockopt: $!";
bind (Server, sockaddr_in ($port, INADDR_ANY)) or die "bind: $!";
listen (Server, SOMAXCONN) or die "listen: $!";

logmsg "server started on port $port";
$SIG{INT} = \&catch_int;

# TODO: timeout on client connections

for (; $paddr = accept (Client, Server); close Client) {
	my($port,$iaddr) = sockaddr_in ($paddr);
	my $name = gethostbyaddr ($iaddr, AF_INET);

	$client_name = "" . inet_ntoa ($iaddr) . ":$port";
	logmsg "cddb< open from $name [$client_name]";
	eval {cddb_server ();};
	logmsg "cddb> close $name [$client_name]";
}

sub catch_int {
	logmsg "server shutdown";
	exit;
}

sub client {
	logmsg "cddb> @_" if ($TRACES =~ /PROXY/);
	print Client "@_\r\n";
}

sub cddb_server {
	local $name, $host, $tool, $version;
	local $proto;

	select (Client);
	$| = 1;
	select (STDOUT);

	undef $hello;
	client "201 $hostname CDDB Proxy $version ready at " .  scalar localtime ();

	while (<Client>) {
		next if (/^\s+$/);

		s/[\n\r]//g;
		logmsg "cddb< $_" if ($TRACES =~ /CLIENT/);

		if (/quit/) {
			last;
		} elsif (/proto/) {
			if ($' =~ /(\d+)/) {
				if ($1 eq $proto) {
					client "502 Protocol level already $proto.";
					next;
				}
				$proto = $1;
				if ($proto > $supported_cddb) {
					client "501 Illegal CDDB protocol level.";
					next;
				}
				client "201 OK, CDDB protocol level now: $proto";
				next;
			}
                  client "500 Command syntax error.";
                  next;
		} elsif (/cddb hello/) {
			if ($hello) {
				client "402 Already shook hands";
				next;
			} elsif ($' =~ /(\S+) (\S+) (\S+) (\S+)/) {
		            $name = $1;
       		      $host = $2;
       		      $tool = $3;
               		$version = $4;

				if ("$1$2$3$4" =~ /^\s+$/) {
					client "431 Handshake not successful, closing connection";
					exit;
				}
                		$hello = true;
                		client "200 Hello and welcome $name\@$host running $tool $version";
				next;
 			} 
			client "500 Command syntax error.";
			next;
		} elsif (/cddb\s+/) {
			if (!$hello) {
				client "409 No handshake.";
				next;
			}
			handle_cddb ($');
			next;
		}

		client "500 Unrecognized command.";
	}
	
	client "230 $hostname Closing connection.  Goodbye.";
}

sub handle_cddb {
	my($cmd) = @_;
	my $in_local;
	my @slurp;

	$cmd =~ s/\s/+/g;

	if ($ENABLE_CACHE eq true) {
		if ($cmd =~ /^query\+([0-9a-f]+)/) {
			logmsg "cddb< query on id $1";
			return if (query_local ($cmd) ne false);
		} elsif ($cmd =~ /^read\+(\S+)\+([0-9a-f]+)$/) {
			logmsg "cddb< read for id $2 ($1)";
			return if (read_local ($cmd) ne false);
			$in_local = false;
		}
	}

# TODO: check for net connection somehow?
	if ($SPOOL eq true) {
		if ($cmd =~ /^lscat/) {
			chdir $LOCAL_CDDB || die "chdir: $!";
			opendir DIR, "." || die "opendir: $!";
			my @all_cats = grep !/^\.\.?$/, grep -d, readdir DIR;
			closedir DIR;

			client "210 Okay category list follows (until terminating market)";
			map { client $_; } @all_cats;
			client ".";
		}
		return;
	}

	open_http ($cmd);
	my $state = 0;
	while (<HTTP>) {
		s/[\n\r]//g;
		logmsg "http< $_" if ($TRACES =~ /HTTP/);
		if (/^\s*$/) {
			$state++;
			next;
		}
		if ($state > 0) {
			client $_;
			push @slurp, $_ if ($in_local eq false);
		}
	}
	close HTTP;
	update_local ($cmd, @slurp) if ($in_local eq false);
	return;

	client "500 Command syntax error.";
}

# TODO: timeout on http connection

sub open_http {
	my($cmd) = @_;

	my @remote = gethostbyname ($HTTP_HOST);
	my $rhost = @remote[0] || $HTTP_HOST;
	my $proto  = getprotobyname ('tcp');

	my $port;
	if ($HTTP_PORT =~ /(\d+)/) {
		$port = $1;
	} else {
		$port  = getservbyname ($HTTP_PORT, 'tcp');
	}

	logmsg "http> open to $rhost:$port";

	socket (HTTP, PF_INET, SOCK_STREAM, $proto) or die "socket: $!";
	my $iaddr = inet_aton ($rhost) or die "inet_aton: $!";
	my $paddr = sockaddr_in ($port, $iaddr) or die "inet_aton: $!";
	connect (HTTP, $paddr) or die "connect: $!";

	my $old = select (HTTP);
	$| = 1;
	select ($old);

	my $ident = "hello=$name+$host+$tool+$version";
	my $proto = "proto=2";
	my $run = "GET ${HTTP_PATH}?cmd=cddb+${cmd}&${ident}&${proto} HTTP/1.0";
	
	logmsg "http> $run";
	print HTTP "$run\r\n\r\n";
}

# cddb query c90d1b0f 15 150 17797 33702 51852 67955 89787 105100 
#    107500 122120 139042 165370 181972 203527 222040 238340 3357

sub query_local {
	my($cmd) = @_;
	my @all_cats;

	$cmd =~ /^query\+([0-9a-f]+)\+(.+)$/;
	my($id) = $1;
	my(@look) = split (/\+/, $2);

# search for disc id
	chdir $LOCAL_CDDB || die "chdir: $!";
	opendir DIR, "." || die "opendir: $!";
	@all_cats = grep !/^\.\.?$/, grep -d, readdir DIR;
	closedir DIR;

	my $key;
	foreach $key (@all_cats) {
		if (-e "$key/$id") {
			my $dtitle;
			open FILE, "<$key/$id" or die "open: $key/$id";
			while (<FILE>) {
				chop;
				if (/^DTITLE=(.*)$/) {
					$dtitle = $1;
					last;
				}
			}
			close FILE;
			client "200 $key $id $dtitle";
			logmsg "cddb> exact match $id ($key) \"$dtitle\" from cache";

			return true;
		}
	}

	logmsg "id $id not in cache";

# check fuzzy values
if ($ENABLE_FUZZY eq true) {
	my @fuzzy;

	chdir $LOCAL_CDDB || die "chdir: $!";
	opendir DIR, "." || die "opendir: $!";
	@all_cats = grep !/^\.\.?$/, grep -d, readdir DIR;
	closedir DIR;

	foreach $key (@all_cats) {
		chdir $key || die "chdir: $!";
		opendir DIR, "." || die "opendir: $!";
		@discs = grep -f, readdir DIR;
		closedir DIR;

		foreach $disc (@discs) {
			my $dtitle = "";
			my $start = false;
			my @track;
			open FILE, "<$disc" or die "open: $disc";
			while (<FILE>) {
				chop;
				if ($start eq false && /^# Track frame offsets:$/i) {
					$start = true;
				} elsif ($start eq true) {
					if (/^#\t(\d+)$/) {
						push @track, $1;
					} elsif (/^DTITLE=(.*)$/) {
						$dtitle = $1;
						last;
					}
				}
			}
			close FILE;
			if (@track eq @look[0]) {
				my $dist = 0;
				for ($i = 1; $i <= @track; $i++) {
					$dist += @track[$i] - @look[$i];
				}
				$dist = ($dist < 0) ? -$dist : $dist;
				push @fuzzy, "$key $disc $dtitle" if ($dist < $tolerance);
			}
			undef @track;
		}
		chdir "..";
	}

	if (@fuzzy > 0) {
		client "211 Found inexact matches, list follows (until terminating marker)";
		map { client $_; } @fuzzy;
		client ".";
		undef @fuzzy;
	}
}
	
# set in queue
	if ($SPOOL eq true) {
		logmsg "creating spool file $CDDB_SPOOL/$id";
		open FILE, ">$CDDB_SPOOL/$id" || die "open: $!";
		print FILE "$cmd\n";
		close FILE;
		print Client "202 No match found.\r\n";
		return true;
	}

	return false;
}

# cddb read rock c90d1b0f
#

sub read_local {
	my($cmd) = @_;

	$cmd =~ /^read\+(\S+)\+([0-9a-f]+)$/;
	my($category) = $1;
	my($id) = $2;
	my($search) = "$LOCAL_CDDB/$category/$id";

	if (!-e $search) {
		print Client "401 Specified CDDB entry not found.\r\n" if ($SPOOL eq true);
		logmsg "cache miss ($id)";
		return false;
	}

	if (-z $search) {
		print Client "403 Database entry is corrupt.\r\n" if ($SPOOL eq true);
		logmsg "cache error $search has zero size - removing file";
		unlink $search;
		return false;
	}

	logmsg "cddb> sending entry $id from cache";
	print Client "210 OK, CDDB proxy entry follows (until terminating marker)\r\n";
	open FILE, $search || die "open: $!";
	while (<FILE>) {
		chop;
		print Client "$_\r\n";
		print "$_\n" if ($TRACES =~ /CACHE/);
	}
	close FILE;
	print Client ".\r\n";
	print ".\n" if ($TRACES =~ /CACHE/);

	return true;
}

# update from a successful response from a cddb read
#

sub update_local {
	my($cmd, @msg) = @_;

	$cmd =~ /^read\+(\S+)\+([0-9a-f]+)$/;
	my($category) = $1;
	my($id) = $2;
	my($search) = "$LOCAL_CDDB/$category";

	if (-e "$search/$id") {
		logmsg "warning: file $search/$id exists";
		return;
	}

	if (not -e $search) {
		logmsg "warning: category directory doesn't exist - creating $search";
		$search =~ tr/[A-Z]/[a-z]/ if ($LOWER_CASE_CDDB eq true);
		mkdir $search, 0755;
	}

	my $key;
	$key = shift @msg;
	if ($key =~ /^210 /) {
		logmsg "storing entry $search/$id";
		open FILE, ">$search/$id" || die "open: $!";
		foreach $key (@msg) {
			last if ($key =~ /^\./);
			print FILE "$key\n";	
		}
		close FILE;
	} else {
		logmsg "not storing: code is \"$key\"";
	}
}

# trawl thru spool directory submit files to remote cddb and
# update local cache with entries

sub clean_spool {
	my @all;

	chdir $CDDB_SPOOL || die "chdir: $!";
	opendir DIR, "." || die "opendir: $!";
	@all = grep -f, readdir DIR;
	closedir DIR;

	local $name = "cddbq";
	local $host = "127.0.0.1";
	local $tool = "CDDBQ";

	logmsg "checking spool directory $CDDB_SPOOL";

	foreach my $key (@all) {
		logmsg "reading spool file $key";

# read file
		open FILE, $key || die "open: $!";
		my $query = <FILE>;
		close FILE;
		
# submit query over http
		$query =~ s/[\n\r]//g;
		open_http ($query);

# parse results
		my $state = 0;	
		my @slurp;
		while (<HTTP>) {
			s/[\n\r]//g;
			logmsg "http: $_" if ($TRACES =~ /HTTP/);
			if (/^\s*$/) {
				$state++;
				next;
			}
			if ($state > 0) {
				push @slurp, $_;
			}
		}
		close HTTP;

# submit read for each result
		my $entry = @slurp[0];
		my @cmd_queue;
		my $code, $cat, $id, $title;

		$entry =~ /^(\d+)/;
		$code = $1;

# exact match
# 200 rock 980bfe0c Meredith Brooks / Blurring The Edges
		if ($code eq 200) {
			$entry =~ /(\d+) (\S+) ([0-9a-f]+) (.+)/;
			$cmd = "read+$2+$3";
			push @cmd_queue, $cmd;

# inexact match
# 211 Found inexact matches, list follows (until terminating `.')
# misc 9a10b40d Britney Spears / Baby one more time
# .
#
# or
# 210 Found exact matches, list follows (until terminating marker)
# .
		} elsif ($code eq 211 || $code eq 210) {
			pop @slurp;
			my $line;
			while ($line = pop @slurp) {
				if ($line =~ /211 /) {
					last;
				} elsif ($line =~ /(\S+) ([0-9a-f]+) (.+)/) {
					$cmd = "read+$1+$2";
					logmsg "queuing id $1/$2";
					push @cmd_queue, $cmd;
				}
			}

# not found
# server error
# etc...
		} else {
			logmsg "error with track $id - return code $code";
			next; # do not remove spool file
		}

		foreach $cmd (@cmd_queue) {
			$cmd =~ /(\S+)\+(\S+)\+([0-9a-f]+)/;
			logmsg "requesting further details for id $2/$3";
			open_http ($cmd);
			my $state = 0;
			my @slurp;
			while (<HTTP>) {
				s/[\n\r]//g;
				logmsg "http: $_" if ($TRACES =~ /HTTP/);
				if (/^\s*$/) {
					$state++;
					next;
				}
				if ($state > 0) {
					push @slurp, $_;
				}
			}
			close HTTP;

# store result in local cddb
			update_local ($cmd, @slurp);
		}

# done with spooled key, remove file

		logmsg "removing spool file $CDDB_SPOOL/$key";
		unlink $key;
	}
}

# eof

